.notification-box {
  background-color: lightgray;
  padding: 10px 12px 10px 10px;
  border-radius: 6px;
  display: flex;
  align-items: center;
  cursor: default;
}


.notification-box-success {
  background-color: #F9FCF6;
  border: 1px solid #E0FACD;
  border-left: 4px solid #91B974;
}

.notification-box-success .notification-box-icon {
  color: #91B974;
}

.notification-box-info {
  background-color: #F6F9FC;
  border: 1px solid #CDE2FA;
  border-left: 4px solid #3B96FF;
}

.notification-box-info .notification-box-icon {
  color: #3B96FFFF;
}

.notification-box-warning {
  background-color: #fcf8e3;
  border: 1px solid #faebcc;
  border-left: 4px solid #cd9c38;
}

.notification-box-warning .notification-box-icon {
  color: #cd9c38;
}


.notification-box-error {
  background-color: #FCF7F6;
  border: 1px solid #FAD1CD;
  border-left: 4px solid #FF6458;
}

.notification-box-error .notification-box-icon {
  color: #FF6458;
}


.notification-box-icon {
  font-size: 20px;
  margin-right: 15px;
}

.notification-box-title {
  font-size: 14px;
  font-weight: bold;
}
